/*
 * Title:        GridSim Toolkit
 * Description:  GridSim (Grid Simulation) Toolkit for Modeling and Simulation
 *               of Parallel and Distributed Systems such as Clusters and Grids
 * Licence:      GPL - http://www.gnu.org/copyleft/gpl.html
 *
 * $Id: GridSimCore.java,v 1.5 2004/05/29 04:40:50 anthony Exp $
 */

package gridsim;

import eduni.simjava.*;
import gridsim.Input;
import gridsim.Output;

/**
 * Since GridSim version 3.0, this is the overall class of GridSim package,
 * that must be extended by other GridSim
 * entities. It inherits event management and threaded entity features from
 * the <b>eduni.simjava.Sim_entity</b> class. This class adds networking and
 * event delivery features, which allow synchronous or asynchronous
 * communication for service access or delivery.
 * <p>
 * All classes that extend this class can implement a method called
 * <tt>body()</tt>, which is automatically invoked since it is expected to be
 * responsible for simulating entity behavior. <tt>body()</tt> method is
 * the place to receive incoming events.
 * <p>
 * The entities that extend this class can be instantiated <i>with</i> or
 * <i>without</i> networked I/O ports. A networked GridSim entity gains
 * communication capability via the objects of GridSim's I/O entity classes,
 * {@link gridsim.Input} and {@link gridsim.Output} classes. Each I/O entity
 * will have a unique name assuming each GridSim entity that the user creates
 * has a unique name. For example, a resource entity with the name
 * <b>Resource2</b>,
 * will have an input entity whose name is prefixed with <b>Input_</b>,
 * making the input entity's full name <b>Input_Resource2</b>, which is
 * expected to be unique. <b>Input_Resource2</b> entity handles all incoming
 * events. A resource has an output entity whose name is prefixed with
 * <b>Output_</b>, e.g. <b>Output_Resource2</b>. <b>Output_Resource2</b>
 * entity handles all outgoing events.
 * <p>
 * The I/O entities are concurrent entities, but they
 * are visible within the GridSim entity and are able to communicate with
 * other GridSim entities by sending messages.
 * <p>
 * There are two ways to send a message or object to an entity, i.e. with or
 * without I/O port. Below shows the differences:
 * <ul>
 *      <li>using {@link #send(int, double, int, Object)} method. <br>
 *          For example: <tt>send(destID, delay, tag, obj)</tt>. This sends
 *          to <tt>destID</tt> <b>WITHOUT</b> I/O port.<br>
 *          On the destinated entity, by creating a Sim_event object:
 *          <ul>
 *              <li> Sim_event.get_data() is obj of type Object. NOTE:
 *                   Object can be a generic type, such as IO_data, Gridlet,
 *                   String, etc.
 *              <li> Sim_event.get_dest() is destID of type int.
 *              <li> Sim_event.get_src() is sender ID of type int. NOTE:
 *                   sender ID is the entity ID that uses/calls this method.
 *              <li> Sim_event.get_tag() is tag of type int.
 *              <li> Sim_event.get_scheduled_by() is sender ID of type int.
 *          </ul>
 *      <br>
 *      <li>using {@link #send(Sim_port, double, int, Object)} method.<br>
 *          For example: <tt>send(ouputPort, delay, tag, obj)</tt> where
 *          obj <b>MUST</b> be of type <tt>IO_data</tt>. This sends
 *          to <tt>destID</tt> <b>WITH</b> I/O port.<br>
 *          On the destinated entity, by creating a Sim_event object:
 *          <ul>
 *              <li> Sim_event.get_data() is (IO_data) obj.getData() of type
 *                   Object. NOTE:
 *                   Object can be a generic type, such as Gridlet, String,
 *                   Integer, etc.
 *              <li> Sim_event.get_dest() is (IO_data) obj.getDestID() of
 *                   type int.
 *              <li> Sim_event.get_src() is an entity ID of type int.
 *                   <b>NOTE:</b> entity ID is not from the sender ID, but from
 *                   <tt>Input_xxx</tt>, where xxx = destinated entity name.
 *                   For example, to send to <tt>Resource2</tt>, the entity ID
 *                   is an ID of <tt>Input_Resource2</tt>.
 *              <li> Sim_event.get_tag() is tag of type int.
 *              <li> Sim_event.get_scheduled_by() is an entity ID of type int.
 *                   Same explanation of Sim_event.get_src().
 *          </ul>
 *          <br>
 *          <b>NOTE:</b> When sending using I/O port, object must be of type
 *             <tt>IO_data</tt>, otherwise ClassCastException will be reported
 *             on {@link gridsim.Output#body()} method.
 *
 * </ul>
 *
 * @author       Manzur Murshed and Rajkumar Buyya
 * @version      3.0, May 2004
 * @since        GridSim Toolkit 3.0
 * @see eduni.simjava.Sim_entity
 * @see gridsim.Output
 * @see gridsim.Input
 * @invariant $none
 */
public class GridSimCore extends Sim_entity
{
    private boolean networkedFlag_;  // true, if networked entity, other false

    // false means NOT invoked
    private boolean terminateIOEntitiesFlag_ = false;

    /** Reading data received via input port */
    protected Sim_port input;

    /** Sending data via output port to external entities */
    protected Sim_port output;



    /**
     * Allocates a new GridSim object
     * <b>without</b> NETWORK communication channels: "input" and
     * "output" Sim_port. In summary, this object has <tt>NO</tt>
     * network communication or bandwidth speed.
     * @param name       the name to be associated with this entity (as
     *                   required by Sim_entity class from simjava package)
     * @throws Exception This happens when creating this entity before
     *                   initializing GridSim package or this entity name is
     *                   <tt>null</tt> or empty
     * @see gridsim.GridSim#init(int, Calendar, boolean, String[], String[],
     *          String)
     * @see eduni.simjava.Sim_entity
     * @pre name != null
     * @post $none
     */
    public GridSimCore(String name) throws Exception
    {
        super(name);
        networkedFlag_ = false;
        input = null;
        output = null;
    }

    /**
     * Allocates a new GridSim object
     * <b>with</b> NETWORK communication channels: "input" and
     * "output" Sim_port. In addition, this method will create <tt>Input</tt>
     * and <tt>Output</tt> object.
     * @param name       the name to be associated with this entity (as
     *                   required by Sim_entity class from simjava package)
     * @param baudRate   network communication or bandwidth speed
     * @throws Exception This happens when creating this entity before
     *                   initializing GridSim package or this entity name is
     *                   <tt>null</tt> or empty
     * @see gridsim.GridSim#init(int, Calendar, boolean, String[], String[],
     *          String)
     * @see eduni.simjava.Sim_entity
     * @see gridsim.Input
     * @see gridsim.Output
     * @pre name != null
     * @pre baudRate > 0.0
     * @post $none
     */
    public GridSimCore(String name, double baudRate) throws Exception
    {
        super(name);
        networkedFlag_ = true;

        input = new Sim_port("input");
        output = new Sim_port("output");

        super.add_port(input);
        super.add_port(output);

        // Every GridSim entity with network has its own input/output channels.
        // Connect this entity "input" port to its input buffer "in_port"
        new Input("Input_" + name, baudRate);
        Sim_system.link_ports(name, "input", "Input_" + name, "input_buffer");

        new Output("Output_" + name, baudRate);
        Sim_system.link_ports(name, "output", "Output_"+name, "output_buffer");
    }

    /**
     * It terminates Entities managing NETWORK communication channels.
     * It can be invoked explicity to shutdown NETWORK communication channels.
     * It is advisable for all entities extending GridSim class, explicitly
     * invoke this method to terminate <tt>Input</tt> and
     * <tt>Output</tt> entities created
     * by the constructor: {@link GridSim#GridSim(String, double)}
     * @deprecated As of GridSim 2.1, replaced by
     *             {@link #terminateIOEntities()}
     * @pre $none
     * @post $none
     */
    protected void TerminateInputOutputEntities() {
        terminateIOEntities();
    }

    /**
     * It terminates Entities managing NETWORK communication channels.
     * It can be invoked explicity to shutdown NETWORK communication channels.
     * It is advisable for all entities extending GridSim class, explicitly
     * invoke this method to terminate <tt>Input</tt> and
     * <tt>Output</tt> entities created
     * by the constructor: {@link GridSim#GridSim(String, double)}
     * @pre $none
     * @post $none
     */
    protected void terminateIOEntities()
    {
        // If it is Networked entity and Not yet terminated, then terminate.
        if ( isNetworked() && !terminateIOEntitiesFlag_ )
        {
            // Send END_OF_SIMULATION to Input entity
            send(input, 0.0, GridSimTags.END_OF_SIMULATION);

            // Send END_OF_SIMULATION to Output entity
            send(output, 0.0, GridSimTags.END_OF_SIMULATION);

            terminateIOEntitiesFlag_ = true;
        }
    }

    /**
     * It terminates the entities of this object that manage <tt>NETWORK</tt>
     * communication channels
     * @deprecated As of GridSim 2.1, replaced by
     *      {@link #finalizeGridSimulation()}
     * @see gridsim.GridSim#terminateIOEntities()
     * @pre $none
     * @post $none
     */
    protected void finalize() {
        finalizeGridSimulation();
    }

    /**
     * It terminates the entities of this object that manage <tt>NETWORK</tt>
     * communication channels
     * @see gridsim.GridSim#terminateIOEntities()
     * @pre $none
     * @post $none
     */
    protected void finalizeGridSimulation() {
        terminateIOEntities();
    }

    /**
     * Check type of entity
     * @return true if entity has NETWORK communication channel, otherwise
     *          it returns false
     * @deprecated As of GridSim 2.1, replaced by {@link #isNetworked()}
     * @pre $none
     * @post $result == true || false
     */
    public boolean IsNetworked() {
        return isNetworked();
    }

    /**
     * Check type of entity
     * @return true if entity has NETWORK communication channel, otherwise
     *          it returns false
     * @pre $none
     * @post $result == true || false
     */
    public boolean isNetworked() {
        return networkedFlag_;
    }


    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param entityName the name of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time.
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @deprecated As of GridSim 2.1, replaced by {@link #send(String, double,
     *             int)}
     * @pre entityName != null
     * @pre delay >= 0.0
     * @post $none
     */
    protected void Send(String entityName, double delay, int gridSimTag) {
        send(entityName, delay, gridSimTag);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param entityName the name of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @pre entityName != null
     * @pre delay >= 0.0
     * @post $none
     */
    protected void send(String entityName, double delay, int gridSimTag)
    {
        if (entityName == null) {
            return;
        }

        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(GridSim.getEntityId(entityName), delay, gridSimTag);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param entityName the name of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @deprecated As of GridSim 2.1, replaced by {@link #send(String, double,
     *             int, Object)}
     * @pre entityName != null
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void Send(String entityName, double delay, int gridSimTag,
            Object data)
    {
        send(entityName, delay, gridSimTag, data);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param entityName the name of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @pre entityName != null
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void send(String entityName, double delay, int gridSimTag,
            Object data)
    {
        if (entityName == null) {
            return;
        }

        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(GridSim.getEntityId(entityName), delay, gridSimTag,
                data);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param entityID   the id number of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @deprecated As of GridSim 2.1, replaced by {@link #send(int, double,
     *             int)}
     * @pre entityID > 0
     * @pre delay >= 0.0
     * @post $none
     */
    protected void Send(int entityID, double delay, int gridSimTag) {
        send(entityID, delay, gridSimTag);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param entityID   the id number of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @pre entityID > 0
     * @pre delay >= 0.0
     * @post $none
     */
    protected void send(int entityID, double delay, int gridSimTag)
    {
        if (entityID < 0) {
            return;
        }

        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(entityID, delay, gridSimTag);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param entityID   the id number of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @deprecated As of GridSim 2.1, replaced by {@link #send(int, double,
     *             int, Object)}
     * @pre entityID > 0
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void Send(int entityID, double delay, int gridSimTag, Object data)
    {
        send(entityID, delay, gridSimTag, data);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param entityID   the id number of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @pre entityID > 0
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void send(int entityID, double delay, int gridSimTag, Object data)
    {
        if (entityID < 0) {
            return;
        }

        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(entityID, delay, gridSimTag, data);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param destPort   A reference to the port to send the event out of
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @deprecated As of GridSim 2.1, replaced by {@link #send(Sim_port, double,
     *             int)}
     * @pre destPort != null
     * @pre delay >= 0.0
     * @post $none
     */
    protected void Send(Sim_port destPort, double delay, int gridSimTag) {
        send(destPort, delay, gridSimTag);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param destPort   A reference to the port to send the event out of
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @pre destPort != null
     * @pre delay >= 0.0
     * @post $none
     */
    protected void send(Sim_port destPort, double delay, int gridSimTag)
    {
        if (destPort == null) {
            return;
        }

        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(destPort, delay, gridSimTag);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param destPort  A reference to the port to send the event out of
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       an object of type IO_data
     * @deprecated As of GridSim 2.1, replaced by {@link #send(Sim_port, double,
     *             int, Object)}
     * @pre destPort != null
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     * @see gridsim.IO_data
     */
    protected void Send(Sim_port destPort, double delay, int gridSimTag,
            Object data)
    {
        send(destPort, delay, gridSimTag, data);
    }

    /**
     * Sends an event/message to another entity by <tt>delaying</tt>
     * the simulation time
     * from the current time, with a tag representing the event type.
     * <p>
     * It is <tt>recommended</tt> to use <tt>send()</tt> method with
     * an output port if the network bandwidth plays an important role in
     * this simulation. However, the entity must have the network entities, i.e.
     * <tt>Input</tt> and <tt>Output</tt> port (specified during the creation
     * of the entity by giving a baud rate or bandwidth speed).
     * Below is an example on how to do:
     * <p>
     * <code>
     * ... // other code <br>
     * // object is the entity or message you want to send <br>
     * // size is the object size in bytes (rough estimation) <br>
     * // destination id is the entity ID you want to send the object to <br>
     * IO_data data = new IO_data(object, size, destinationID); <br>
     * <br> <br>
     * // If this entity extends from GridSim class, then you should use it <br>
     * // otherwise need to create a new <tt>Sim_port</tt> object <br>
     * Sim_port port = super.output; <br> <br>
     * // delay is the simulation time delay <br>
     * // tag is the event type (user-defined or choose one from
     *    <tt>GridSimTags</tt> class)<br>
     * send(port, delay, tag, data);<br>
     * ... // remaining other code <br>
     * </code>
     *
     * @param destPort  A reference to the port to send the event out of
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     *                   If delay is a negative number, then it will be
     *                   changed to 0.0
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       an object of type IO_data
     * @pre destPort != null
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     * @see gridsim.IO_data
     */
    protected void send(Sim_port destPort, double delay, int gridSimTag,
            Object data)
    {
        if (destPort == null) {
            return;
        }

        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(destPort, delay, gridSimTag, data);
    }


} // end class

